// Copyright © 2020 Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.register.provider.aws.action;


import com.amazonaws.services.dynamodbv2.datamodeling.DynamoDBDeleteExpression;
import com.amazonaws.services.dynamodbv2.model.AttributeValue;
import com.amazonaws.services.dynamodbv2.model.ConditionalCheckFailedException;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperFactory;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperV2;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.action.model.Action;
import org.opengroup.osdu.register.provider.aws.util.DocUtil;
import org.opengroup.osdu.register.provider.interfaces.action.IActionRepo;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Repository;

import javax.inject.Inject;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Repository
public class AwsActionRepo implements IActionRepo {


    @Autowired
    private DpsHeaders headers;

    @Autowired
    private JaxRsDpsLog logger;

    @Inject
    private DynamoDBQueryHelperFactory dynamoDBQueryHelperFactory;

    @Value("${aws.dynamodb.actionTable.ssm.relativePath}")
    String actionTableParameterRelativePath;

    private DynamoDBQueryHelperV2 getActionQueryHelper() {
        return dynamoDBQueryHelperFactory.getQueryHelperForPartition(headers, actionTableParameterRelativePath);
    }

    @Override
    public List<Action> getAllActions() {
        DynamoDBQueryHelperV2 queryHelper = getActionQueryHelper();
        List<String> filters = new ArrayList<>();
        Map<String, AttributeValue> valueMap = new HashMap<>();

        filters.add("dataPartitionId = :dataPartitionId");
        valueMap.put(":dataPartitionId", new AttributeValue().withS(headers.getPartitionId()));

        String filterExpression = String.join(" and ", filters);
        logger.info(String.format("Action query filter expression: %s", filterExpression));

        List<ActionDoc> results = queryHelper.scanTable(ActionDoc.class, filterExpression, valueMap);

        return results.stream().map(ActionDoc::mapto).collect(Collectors.toList());
    }

    @Override
    public Action createAction(Action action) {
        DynamoDBQueryHelperV2 queryHelper = getActionQueryHelper();
        if (action.getId() == null) {
            logger.error("Action id cannot be null");
            throw new AppException(400, "Bad Request", "Action id cannot be null");
        }

        ActionDoc doc = ActionDoc.mapFrom(action, headers.getPartitionId());

        try {
            queryHelper.saveWithHashCondition(doc, DocUtil.getHashKey());
        } catch (ConditionalCheckFailedException e) {
            logger.error(String.format("An action already exists with the id: %s", action.getId()));
            throw new AppException(409, "Conflict", String.format("An action already exists with the id: %s", action.getId()));
        } catch (AppException e) {
            throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
        }

        return action;
    }


    @Override
    public boolean delete(String id) {
        DynamoDBQueryHelperV2 queryHelper = getActionQueryHelper();
        try {
            ActionDoc objectToDelete = new ActionDoc();
            objectToDelete.setId(id);
            DynamoDBDeleteExpression deleteExpression = new DynamoDBDeleteExpression()
                    .withConditionExpression("attribute_exists(id)");
            queryHelper.deleteByObjectWithCondition(objectToDelete, deleteExpression);

        } catch (ConditionalCheckFailedException e) {
            logger.error(String.format("Failed to delete: %s Object does not exist", id));
            throw new AppException(404, "ObjectNotFound", String.format("The Action you are trying to delete: %s does not exist", id));
        } catch (AppException e) {
            throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
        }
        return true;

    }

    @Override
    public Action get(String id) {
        DynamoDBQueryHelperV2 queryHelper = getActionQueryHelper();
        ActionDoc doc = queryHelper.loadByPrimaryKey(ActionDoc.class, id);
        if (doc == null) {
            logger.error(String.format("Action with id %s does not exist.", id));
            throw new AppException(404, "Not found", String.format("Action with id %s does not exist.", id));
        } else {
            return ActionDoc.mapto(doc);
        }
    }

}
