// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.subscriber;

import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.jupiter.MockitoExtension;
import org.opengroup.osdu.azure.cosmosdb.CosmosStore;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.provider.azure.di.AzureBootstrapConfig;
import org.opengroup.osdu.register.provider.azure.di.CosmosContainerConfig;
import org.opengroup.osdu.register.provider.azure.subscriber.pubsub.IPubsubPartitionHelper;
import org.opengroup.osdu.register.provider.azure.util.CryptographyUtil;
import org.opengroup.osdu.register.subscriber.model.*;

import java.sql.Timestamp;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.*;

@ExtendWith(MockitoExtension.class)
public class CosmosSubscriptionTest {
    private static final String dataPartitionId = "data-partition-id";
    private static final String subscriptionId = "id-1";
    private static final String subscriptionContainer = "ddms-container";
    private static final String cosmosDatabase = "cosmos-database";
    private static final String secretValue = "secret-value";
    private static final String secretType = "secret-type";
    private static final String encryptedData = "some-string";
    private static final Timestamp currentTime = new Timestamp(System.currentTimeMillis());
    private static final String topicName = "topic-name";

    @Mock
    private AzureBootstrapConfig azureBootstrapConfig;

    @Mock
    private CosmosContainerConfig cosmosContainerConfig;

    @Mock
    private CosmosStore cosmosStore;

    @Mock
    private DpsHeaders dpsHeaders;

    @Mock
    private JaxRsDpsLog logger;

    @Mock
    private CryptographyUtil cryptographyUtil;

    @Mock
    private IPubsubPartitionHelper partitionHelper;

    @Mock
    private Subscription subscription;

    @Mock
    private SubscriptionDoc subscriptionDoc;

    @InjectMocks
    private CosmosSubscription sut;

    @BeforeEach
    void init() {
        lenient().doReturn(dataPartitionId).when(dpsHeaders).getPartitionId();
        lenient().when(azureBootstrapConfig.getCosmosDBName()).thenReturn(cosmosDatabase);
        lenient().when(cosmosContainerConfig.getSubscriptionContainerName()).thenReturn(subscriptionContainer);
        lenient().when(subscription.getId()).thenReturn(subscriptionId);
        lenient().when(subscription.getSecret()).thenReturn(new HmacSecret(secretValue));
        lenient().when(subscription.getNotificationId()).thenReturn(subscriptionId);
        lenient().when(subscription.getTopic()).thenReturn(topicName);
        lenient().when(partitionHelper.getFullyQualifiedTopicName(topicName)).thenReturn(topicName);
        lenient().when(partitionHelper.getTopicAliasFromFullyQualifiedTopicName(topicName)).thenReturn(topicName);
        lenient().when(subscriptionDoc.getSecretType()).thenReturn(secretType);
        lenient().when(subscriptionDoc.getSecretValue()).thenReturn(secretValue);
        lenient().when(subscriptionDoc.getCreatedOnEpoch()).thenReturn(currentTime);
        lenient().when(subscriptionDoc.getNotificationId()).thenReturn(subscriptionId);
        lenient().when(subscriptionDoc.getTopic()).thenReturn(topicName);
        lenient().when(cryptographyUtil.encryptData(secretValue)).thenReturn(encryptedData);
        lenient().when(cryptographyUtil.decryptData(encryptedData)).thenReturn(secretValue);
    }

    @Test
    public void shouldCreateCosmosSubscription() {
        doNothing().when(cosmosStore).createItem(eq(dataPartitionId), eq(cosmosDatabase), eq(subscriptionContainer), eq(dataPartitionId), any());
        Subscription output = sut.createSubscription(subscription);
        Assertions.assertEquals(topicName, output.getTopic());
        Assertions.assertEquals("id-1", output.getNotificationId());
        Assertions.assertEquals("HMAC", output.getSecret().getSecretType());
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(cosmosContainerConfig, times(1)).getSubscriptionContainerName();
        verify(dpsHeaders, times(3)).getPartitionId();
        verify(cryptographyUtil, times(1)).encryptData(secretValue);
        verify(cryptographyUtil, times(1)).getKeyName();
        verify(cryptographyUtil, times(1)).getKeyVersion();
        verify(subscription, times(3)).getSecret();
        verify(subscription, times(2)).getId();
        verifyCallsToSubscriptionMock();

    }

    @Test
    public void shouldThrows409WhenIdExistsCosmosSubscription() {
        doThrow(new AppException(409, "Reason", "Message")).when(cosmosStore).createItem(eq(dataPartitionId), eq(cosmosDatabase), eq(subscriptionContainer), eq(dataPartitionId), any());
        AppException exception = assertThrows(AppException.class, () -> {
            sut.createSubscription(subscription);
        });
        Assertions.assertNotNull(exception);
        Assertions.assertEquals(409,exception.getError().getCode());
        verify(cryptographyUtil, times(1)).getKeyName();
        verify(cryptographyUtil, times(1)).getKeyVersion();
        verify(cosmosContainerConfig, times(1)).getSubscriptionContainerName();
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(subscription, times(1)).getId();
        verify(dpsHeaders, times(3)).getPartitionId();
    }

    @Test
    public void shouldDeleteSubscription() {
        doNothing().when(cosmosStore).deleteItem(eq(dataPartitionId), eq(cosmosDatabase), eq(subscriptionContainer), eq(subscriptionId), eq(dataPartitionId));
        sut.deleteSubscription(subscriptionId);
        verify(logger, times(1)).debug("Record deleted for subscription with ID: " + subscriptionId);
    }
    @Test
    public void shouldThrow500ExceptionWhenDeletionFailedCosmosDb() {
        doThrow(new AppException(500, "Server Error", "Unexpected error deleting subscription")).when(cosmosStore).deleteItem(eq(dataPartitionId), eq(cosmosDatabase), eq(subscriptionContainer), eq(subscriptionId), eq(dataPartitionId));
        AppException exception = assertThrows(AppException.class, () -> {
            sut.deleteSubscription(subscriptionId);
        });
        Assertions.assertNotNull(exception);
        Assertions.assertEquals(500, exception.getError().getCode());
        verify(logger, times(0)).info("Record deleted for subscription with ID: " + subscriptionId);
    }

    @Test
    public void shouldGetSubscription() {
        doReturn(Optional.of(subscriptionDoc)).when(cosmosStore).findItem(eq(dataPartitionId), eq(cosmosDatabase), eq(subscriptionContainer), eq(subscriptionId), eq(dataPartitionId), any());
        when(cryptographyUtil.decryptData(secretValue)).thenReturn(secretValue);

        Subscription expected = getSubscription();
        Subscription output = sut.getSubscription(subscriptionId);
        expected.setCreatedOnEpoch(output.getCreatedOnEpoch());

        assertEquals(expected, output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, cosmosDatabase, subscriptionContainer, subscriptionId, dataPartitionId, SubscriptionDoc.class);
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(cosmosContainerConfig, times(1)).getSubscriptionContainerName();
        verify(cryptographyUtil, times(1)).decryptData(secretValue);
        verify(dpsHeaders, times(2)).getPartitionId();
        verifyCallsToSubscriptionDocMock();
    }

    @Test
    public void getSubscriptionItemRetrievedSuccessfullyWithGSASecretType() {
        String audience = "audience";
        String key = "key";
        String gsaSecretStringValue = audience + "`" + key;
        GsaSecretValue gsaSecretValue = new GsaSecretValue(audience, key);
        Subscription expected = getSubscription();
        expected.setSecret(new GsaSecret(gsaSecretValue));

        when(subscriptionDoc.getSecretType()).thenReturn("GSA");
        when(subscriptionDoc.getSecretValue()).thenReturn(gsaSecretStringValue);
        doReturn(Optional.of(subscriptionDoc)).when(cosmosStore).findItem(eq(dataPartitionId), eq(cosmosDatabase), eq(subscriptionContainer), eq(subscriptionId), eq(dataPartitionId), any());
        when(cryptographyUtil.decryptData(gsaSecretStringValue)).thenReturn(gsaSecretStringValue);

        Subscription output = sut.getSubscription(subscriptionId);
        expected.setCreatedOnEpoch(output.getCreatedOnEpoch());

        assertEquals(expected, output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, cosmosDatabase, subscriptionContainer, subscriptionId, dataPartitionId, SubscriptionDoc.class);
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(cosmosContainerConfig, times(1)).getSubscriptionContainerName();
        verify(cryptographyUtil, times(1)).decryptData(gsaSecretStringValue);
        verify(dpsHeaders, times(2)).getPartitionId();
        verifyCallsToSubscriptionDocMock();
    }

    @Test
    public void getSubscriptionThrows404WhenIdNotFound() {
        doReturn(Optional.empty()).when(cosmosStore).findItem(eq(dataPartitionId), eq(cosmosDatabase), eq(subscriptionContainer), eq(subscriptionId), eq(dataPartitionId), any());

        AppException exception = assertThrows(AppException.class, () -> {
            sut.getSubscription(subscriptionId);
        });

        assertEquals(404, exception.getError().getCode());
        assertEquals("Not found", exception.getError().getReason());
        assertEquals("Subscriber with id id-1 does not exist.", exception.getError().getMessage());
        verify(cosmosStore, times(1)).findItem(dataPartitionId, cosmosDatabase, subscriptionContainer, subscriptionId, dataPartitionId, SubscriptionDoc.class);
        verify(cosmosContainerConfig, times(1)).getSubscriptionContainerName();
        verify(dpsHeaders, times(2)).getPartitionId();
        verify(logger, times(1)).error("Subscriber with id id-1 does not exist.");
    }

    @Test
    public void querySubscriptionsReturnItemsSuccessfully() {

        List<SubscriptionDoc> docsList = Collections.singletonList(subscriptionDoc);
        Subscription expected = getSubscription();

        doReturn(docsList).when(cosmosStore).queryItems(eq(dataPartitionId), eq(cosmosDatabase), eq(subscriptionContainer), any(), any(), any());
        when(cryptographyUtil.decryptData(secretValue)).thenReturn(secretValue);

        List<Subscription> output = sut.querySubscription("notification-id");
        expected.setCreatedOnEpoch(output.get(0).getCreatedOnEpoch());

        assertEquals(Collections.singletonList(expected), output);
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(cosmosContainerConfig, times(1)).getSubscriptionContainerName();
        verify(cryptographyUtil, times(1)).decryptData(secretValue);
        verify(dpsHeaders, times(2)).getPartitionId();
        verifyCallsToSubscriptionDocMock();
    }

    @Test
    public void getAllSubscriptionsReturnItemsSuccessfully() {

        List<SubscriptionDoc> docsList = Collections.singletonList(subscriptionDoc);
        Subscription expected = getSubscription();

        doReturn(docsList).when(cosmosStore).queryItems(eq(dataPartitionId), eq(cosmosDatabase), eq(subscriptionContainer), any(), any(), any());
        when(cryptographyUtil.decryptData(secretValue)).thenReturn(secretValue);

        List<Subscription> output = sut.getAllSubscriptions();
        expected.setCreatedOnEpoch(output.get(0).getCreatedOnEpoch());

        assertEquals(Collections.singletonList(expected), output);
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(cosmosContainerConfig, times(1)).getSubscriptionContainerName();
        verify(cryptographyUtil, times(1)).decryptData(secretValue);
        verify(dpsHeaders, times(2)).getPartitionId();
        verifyCallsToSubscriptionDocMock();
    }

    @Test
    public void patchShouldSuccessfullyUpdateSecret() {
        Secret secret = Mockito.mock(Secret.class);
        when(secret.getSecretType()).thenReturn(secretType);
        when(secret.toString()).thenReturn(secretValue);
        when(cryptographyUtil.encryptData(secretValue)).thenReturn(secretValue);
        doNothing().when(cosmosStore).upsertItem(eq(dataPartitionId), eq(cosmosDatabase), eq(subscriptionContainer), eq(dataPartitionId), any());

        boolean output = sut.updateSecret(subscription, secret);

        assertTrue(output);
        verify(secret, times(1)).getSecretType();
        verify(subscription, times(1)).getId();
        verify(subscription, times(2)).getSecret();
        verify(cryptographyUtil, times(1)).encryptData(secretValue);
        verifyCallsToSubscriptionMock();
    }


    private Subscription getSubscription() {
        Subscription expected = new Subscription();
        expected.setSecret(new HmacSecret(secretValue));
        expected.setNotificationId(subscriptionId);
        expected.setTopic(topicName);
        return expected;
    }

    private void verifyCallsToSubscriptionDocMock() {
        verify(subscriptionDoc, times(1)).getSecretType();
        verify(subscriptionDoc, times(1)).getSecretValue();
        verify(subscriptionDoc, times(1)).getId();
        verify(subscriptionDoc, times(1)).getName();
        verify(subscriptionDoc, times(1)).getDescription();
        verify(subscriptionDoc, times(1)).getTopic();
        verify(subscriptionDoc, times(1)).getPushEndpoint();
        verify(subscriptionDoc, times(1)).getCreatedBy();
        verify(subscriptionDoc, times(1)).getCreatedOnEpoch();
        verify(subscriptionDoc, times(1)).getNotificationId();
    }

    private void verifyCallsToSubscriptionMock() {
        verify(subscription, times(1)).getName();
        verify(subscription, times(1)).getDescription();
        verify(subscription, times(1)).getPushEndpoint();
        verify(subscription, times(1)).getCreatedBy();
        verify(subscription, times(1)).getNotificationId();
    }
}
